// For Gifu Challenge. Simple sample code of communication match on nngs.
// WindowsXP, Link wsock32.lib
// 2004/09/16 Hiroshi Yamashita
#include <stdio.h>
#include <time.h>
#include <winsock.h>

char nngs_player_name[2][256] = {
	"playera",	// Black name
	"playerb",	// White name
};
char *stone_str[2] = { "B","W" };

int board[(19+2)*(19+2)];	// Where stone exist is set 1.

SOCKET Sock;

// Send data
void TcpSend(char *ptr)
{
	int n = strlen(ptr);
	send( Sock, ptr, n, 0);
	printf("--->%s",ptr);
}

// Receive data
int TcpRecv(char *ptr)
{
	int sum = 0,nRecv;
	char s[1];

	for (;;) {
		// Read 1byte. until '\n'.
		nRecv = recv( Sock, s, 1, 0);
		if ( nRecv == SOCKET_ERROR ) { printf("recv() Err.\n"); exit(0); }
		*(ptr+sum) = s[0];
		sum++;
		if ( s[0] == '\n' ) break;
		if ( sum > 255 ) { printf("Buff Over!\n"); exit(0); }
	}
	*(ptr+sum) = 0;
	printf("<---%s",ptr);
	return sum;
}

// waiting for server data. check the characters and respond it.
// fTurn .... player color. Black=0, White=1.
int nngsWait(int fTurn)
{
	char str[256];
	char tmp[256];
	char *p;
	int x,y;

	// receive
	for ( ;; ) {
		TcpRecv(str);
		if ( strstr(str,"No Name Go Server") ) {
			TcpSend("set client TRUE\n");	// Change to simple communication mode.
		}

		if ( strstr(str,"Set client to be True") ) {
			if ( fTurn == 0 ) {
				sprintf(tmp,"match %s B 19 40 0\n",nngs_player_name[1-fTurn]);
				TcpSend(tmp);	// If you are Black, you request the match.
			}
		}	

		if ( strstr(str,"Match [19x19]") ) {
			sprintf(tmp,"match %s W 19 40 0\n",nngs_player_name[1-fTurn]);
			TcpSend(tmp);	// If you are White, you receive the match request.
			return -1;
		}

		if ( fTurn == 0 && strstr(str,"accepted.") ) {	// As White accepted the match, you can send first move.
			return -1;
		}	

		if ( strstr(str,"Illegal") ) return -1;		// Error

		sprintf(tmp,"(%s): ",stone_str[1-fTurn]);	// Receive the opponent move
		p = strstr(str,tmp);
		if ( p != NULL ) {		// Receive the usual move
			p += 5;
			if ( strstr(p,"Pass") ) {
				printf("Opponent=PASS!\n");
				return 0;
			} else {
				x = *p;
				y = atoi(p+1);
				printf("Opponent=%c%d\n",x,y);
				x = x  - 'A' - (x > 'I') + 1;	// does not use 'I'
				return y*21 + x;
			}
		}	

		// If two passes, you will receive this message. Server does not send second "Pass".
		if ( strstr(str,"You can check your score") ) {
			sprintf(tmp,"done\n");	// send "done" commmand to end score counting.
			TcpSend(tmp);
			continue;	// wait for "resigns." to finish game. This makes server save complete game SGF file.
 		}
		// "9 {Game 1: test2 vs test1 : Black resigns. W 10.5 B 6.0}"
		if ( strstr(str,"9 {Game") && strstr(str,"resigns.") ) {
			sprintf(tmp,"%s vs %s",nngs_player_name[1],nngs_player_name[0]);
			if ( strstr(str,tmp) ) return -1;	// game end
		}
		
		sprintf(tmp,"{%s has disconnected}",nngs_player_name[1-fTurn]);
		if ( strstr(str,tmp) ) return -1;	// Opponent player has dissconnect from server.

		// player lost on time.
		if ( strstr(str,"forfeits on time") &&
		     strstr(str,nngs_player_name[0]) && strstr(str,nngs_player_name[1]) ) {
			return -1;
		}
	}
}

char sIP[] = "192.168.1.111";	// Local server IP
//char sIP[] = "nngs.computer-go.jp";	// test server on CGF Web.


// connect to nngs
int connect_nngs(void)
{
	WSADATA WsaData;
	SOCKADDR_IN Addr;
	unsigned short nPort = 9696;	// Port number of the server
	unsigned long addrIP;			// proper number of IP address.
	PHOSTENT host = NULL;			// host information
	int i;
	
	// initialize WinSock
	if ( WSAStartup(0x0101, &WsaData) != 0 ) {
		printf("Error!: WSAStartup(): %d\n",WSAGetLastError());
		return 0;
	}

	// make socket
	Sock = socket(PF_INET, SOCK_STREAM, 0);
	if ( Sock == INVALID_SOCKET ) {
		printf("Error: socket(): %d\n", WSAGetLastError());
		return 0;
	}
	
	// get IP from host URL
	addrIP = inet_addr(sIP);
	if ( addrIP == INADDR_NONE ) {
		host = gethostbyname(sIP);		// get IP from host
		if ( host == NULL ) {
			printf("Error: gethostbyname(): \n");
			return 0;
		}
		addrIP = *((unsigned long *)((host->h_addr_list)[0]));
	} else {
//		host = gethostbyaddr((const char *)&addrIP, 4, AF_INET);	// Fail when local IP.
	}

	if ( host ) {	// Dsiplay IP and host.
		printf("public name = %s\n",host->h_name);
		for (i=0; host->h_aliases[i]; i++) printf("alias = %s\n",host->h_aliases[i]);
		for (i=0; host->h_addr_list[i]; i++) {
			IN_ADDR *ip;
			ip = (IN_ADDR *)host->h_addr_list[i];
			printf("IP = %s\n",inet_ntoa(*ip));
		}
	}

	// Set the IP address and Port number of the server.
	memset(&Addr, 0, sizeof(Addr));
	Addr.sin_family      = AF_INET;
	Addr.sin_port        = htons(nPort);
	Addr.sin_addr.s_addr = addrIP;

	// connect the server
	if ( connect( Sock, (LPSOCKADDR)&Addr, sizeof(Addr) ) !=0 ) {
		printf("Error: connect(): %d\n",WSAGetLastError());
		return 0;
	}
	return 1;
}

// close the connect
void close_nngs(void)
{
	closesocket(Sock);
	WSACleanup();
}

// random engine
int thinking(void)
{
	int i,x,y,z;
	
	for (i=0;i<300;i++) {
		x = (rand() % 17) + 2;	// 2-18. does not select board edge.
		y = (rand() % 17) + 2;	
		z = y * 21 + x;
		if ( board[z] != 0 ) continue;
		if ( board[z+1] + board[z-1] + board[z+21] + board[z-21] == 4 ) continue;
		return z;
	}
	return 0;	// Pass
}

// nngs_try W test1 test2
// ... Black is test1, White is test2, login as White.
int main(int argc, char *argv[])
{
	int x,y,z;
	int fTurn = 0;	// I'm Black. (default)
	char str[256];

	// Check the start options
	if ( argc >= 2 && *argv[1] == 'W' ) fTurn = 1; // White
	if ( argc >= 4 ) {
		strcpy(nngs_player_name[0],argv[2]);	// User names
		strcpy(nngs_player_name[1],argv[3]);
	}
	if ( argc == 5 ) strcpy(sIP,argv[4]);		// Server IP

	srand(time(NULL));	// initialize the random seed

	connect_nngs();	// connect to nngs
	
	// Login as the user name.
	sprintf(str,"%s\n",nngs_player_name[fTurn]);
	TcpSend(str);	

	nngsWait(fTurn);	// Set to simple display, respond match request

	if ( fTurn == 0 ) goto black_first;

	for (;;) {
		z = nngsWait(fTurn);
		if ( z < 0 ) break;
		board[z] = 1;

black_first:
		z = thinking();	// random engine

		if ( z == 0 ) sprintf(str,"pass\n");
		else {
			y =  z / 21;
			x =  z - y*21;
			x = 'A' + x - 1;
			if ( x >= 'I' ) x++;	// 'I' is not used. 1 - 19 ... 'A'-'H' and 'J'-'T'
			sprintf(str,"%c%d\n",x,y);
			board[z] = 1;
		}
		TcpSend(str);
	}
	close_nngs();	// close connection.

	printf("Communication finished. Hit enter to close.\n");
	getchar();
}
